class MemoryCache {
    /**
     *
     * @param storage L2 cache
     */
    constructor(storage) {
        this.storage = storage;
        // mem cache
        this.valueByKey = new Map();
    }
    get(key) {
        if (this.valueByKey.has(key)) {
            return Promise.resolve(this.valueByKey.get(key));
        }
        return this.storage.get(key).then((value) => {
            // mem cache version is always newer and should be used
            if (this.valueByKey.has(key)) {
                return this.valueByKey.get(key);
            }
            // load to mem cache
            this.valueByKey.set(key, value);
            return value;
        });
    }
    set(key, value) {
        this.valueByKey.set(key, value);
        return this.storage.set(key, value);
    }
    remove(key) {
        // keep the key in mem cache, key = undefined indicates the key was loaded from L2 cache
        //  so it won't bother to load it again in get()
        this.valueByKey.set(key, undefined);
        return this.storage.remove(key);
    }
    getKeys() {
        return this.storage.getKeys();
    }
    clear() {
        // clear all map keys
        this.valueByKey.forEach((value, key) => {
            if (value === undefined) {
                return;
            }
            this.valueByKey.set(key, undefined);
        });
        return this.storage.clear();
    }
}
export function withMemoryCache(cache) {
    return new MemoryCache(cache);
}
class DummyCache {
    constructor() {
        this.keys = new Set();
    }
    get(key) {
        return Promise.resolve(undefined);
    }
    set(key, value) {
        this.keys.add(key);
        return Promise.resolve();
    }
    remove(key) {
        this.keys.delete(key);
        return Promise.resolve();
    }
    getKeys() {
        return Promise.resolve(Array.from(this.keys));
    }
    clear() {
        this.keys.clear();
        return Promise.resolve();
    }
}
export function createMemoryCache() {
    return withMemoryCache(new DummyCache());
}
