import { CacheMode, BackoffMode, } from './types';
import { createMemoryCache } from './caches/memory';
import { KeyPrefixHelper } from './caches/utils';
import { RequestFactory } from './requests/factory';
const defaultFetcherOptions = {
    cacheMode: CacheMode.Default,
    cacheMaxAge: 3600,
    cacheMinFresh: 1,
};
const defaultRequestOptions = {
    retryTimes: 0,
    retryBackoff: BackoffMode.JitteredExponential,
    retryInitialWaitTime: 1,
};
export class FetcherImpl {
    constructor(requestCreator, options) {
        this.options = Object.assign({}, defaultFetcherOptions, { cache: createMemoryCache() });
        this.ongoingClearCache = Promise.resolve();
        this.config(options);
        this.requestFactory = new RequestFactory(requestCreator);
    }
    config(options) {
        Object.assign(this.options, options);
    }
    clearCache(maxAge, cache) {
        const cacheToClear = cache !== null && cache !== void 0 ? cache : this.options.cache;
        const prefixHelper = new KeyPrefixHelper(this.options.cacheKeyPrefix);
        // serialize the calls to clearCache() and have fetch wait for the clearCache finish
        const thisClearCache = this.ongoingClearCache
            .then(() => cacheToClear.getKeys())
            .then((keys) => {
            return Promise.all(keys
                .filter((key) => prefixHelper.matchPrefix(key))
                .map((key) => {
                if (maxAge > 0) {
                    return cacheToClear.get(key).then((data) => {
                        // cache is valid
                        if ((data === null || data === void 0 ? void 0 : data.timestamp) + maxAge * 1000 > Date.now()) {
                            return;
                        }
                        return cacheToClear.remove(key);
                    });
                }
                return cacheToClear.remove(key);
            }));
        })
            .then(() => {
            if (this.ongoingClearCache === thisClearCache) {
                this.ongoingClearCache = Promise.resolve();
            }
        });
        this.ongoingClearCache = thisClearCache;
        return thisClearCache;
    }
    fetch(request, options) {
        const mergedOptions = Object.assign({}, defaultRequestOptions, this.options, options);
        const fetcherRequest = this.requestFactory.getRequest(mergedOptions, request);
        return {
            abort: () => {
                fetcherRequest.abort();
            },
            response: this.ongoingClearCache.then(() => {
                return fetcherRequest.run();
            }),
        };
    }
}
