import { useEffect, useRef, useCallback } from 'react';
import { AbortErrorName } from '../requests/utils';
import { CacheMode } from '../types';
import { forEachResponse } from '../utils';
import { getURLFetcher } from './fetcher';
import { useDeepEqualMemo, useHookStateRef, useShallowEqualMemo, isEqualForKeys } from './utils';
/**
 * if there are new refresh request comes before previous request finished,
 * keep the one with higher priority (in case the request itself and related options
 * are the same)
 * @param cacheMode
 */
function getSWRCacheModePriority(cacheMode) {
    switch (cacheMode) {
        case CacheMode.NoCache:
            return 4;
        case CacheMode.NoStore:
            return 3;
        case CacheMode.ForceLoad:
            return 2;
        case CacheMode.ForceCache:
            return 1;
        default:
            return 0;
    }
}
// implementation
export function useSWR(fetcherOrURL, request, options) {
    const fetcher = typeof fetcherOrURL === 'string' ? getURLFetcher(fetcherOrURL) : fetcherOrURL;
    const requestMemo = useDeepEqualMemo(request);
    const optionsMemo = useShallowEqualMemo(options);
    const [state, stateRef, updateState, cancelUpdate] = useHookStateRef({
        isLoaded: false,
        isFreshOrValidated: false,
    });
    const abortRef = useRef();
    // keep track of last request & options for compare
    const lastRefreshCacheMode = useRef();
    const lastRequestRef = useRef();
    const lastOptionsRef = useRef();
    const refresh = useCallback((cacheMode) => {
        var _a, _b;
        const mergedOptions = Object.assign({}, optionsMemo);
        if (cacheMode) {
            mergedOptions.cacheMode = cacheMode;
        }
        // reset state to not fully loaded
        updateState({
            isFreshOrValidated: false,
            error: undefined,
            isLoaded: stateRef.current.data !== undefined,
        });
        // check if we need skip this refresh request
        if (abortRef.current) {
            // skip request if there is a ongoing request with the same request and
            // options that covers current refresh request
            if (lastRequestRef.current === requestMemo &&
                isEqualForKeys(lastOptionsRef.current, mergedOptions, [
                    'cache',
                    'cacheKey',
                    'cacheKeyPrefix',
                ]) &&
                getSWRCacheModePriority((_a = lastOptionsRef.current) === null || _a === void 0 ? void 0 : _a.cacheMode) >=
                    getSWRCacheModePriority(mergedOptions.cacheMode)) {
                return;
            }
        }
        // send fetch request
        lastRefreshCacheMode.current = cacheMode;
        lastRequestRef.current = requestMemo;
        lastOptionsRef.current = mergedOptions;
        const thisAbort = forEachResponse(fetcher.fetch(requestMemo, mergedOptions), ({ data, error, next }) => {
            // isFreshOrValidated: no next request
            const isFreshOrValidated = !next;
            // current state
            const currentData = data !== null && data !== void 0 ? data : stateRef.current.data;
            // loaded: any data available or validated
            const isLoaded = currentData !== undefined || isFreshOrValidated;
            // do not handle abort error because
            // it is triggered by client side abort() call or component unmount abort()
            if ((error === null || error === void 0 ? void 0 : error.name) !== AbortErrorName) {
                updateState({
                    data: data !== null && data !== void 0 ? data : stateRef.current.data,
                    error,
                    isLoaded,
                    isFreshOrValidated,
                });
            }
            if (!next && abortRef.current === thisAbort) {
                // clear abort if request fully settled
                abortRef.current = undefined;
            }
        });
        // abort previous request
        (_b = abortRef.current) === null || _b === void 0 ? void 0 : _b.call(abortRef);
        abortRef.current = thisAbort;
    }, [fetcher, updateState, requestMemo, optionsMemo]);
    state.refresh = refresh;
    const abortedOnCleanupRef = useRef(false);
    useEffect(() => {
        // check if there is any pending refresh need to be resume
        if (abortedOnCleanupRef.current) {
            // resume previous pending refresh
            // it was aborted by the cleanup
            refresh(lastRefreshCacheMode.current);
            abortedOnCleanupRef.current = false;
        }
        else {
            // auto start
            const { manualStart } = optionsMemo !== null && optionsMemo !== void 0 ? optionsMemo : {};
            if (!manualStart) {
                refresh();
            }
        }
        return () => {
            // abort fetch if there is any pending request
            if (abortRef.current) {
                abortRef.current();
                abortRef.current = undefined;
                abortedOnCleanupRef.current = true;
            }
            // cancel previous pending state update
            cancelUpdate();
        };
    }, [refresh, cancelUpdate, optionsMemo]);
    return state;
}
