import { useRef, useCallback, useEffect, useReducer } from 'react';
import debounce from 'lodash.debounce';
import isEqual from 'fast-deep-equal';
// Object.is - with +0 -0 case ignored
function is(x, y) {
    if (x === y) {
        return true;
    }
    else {
        return x !== x && y !== y;
    }
}
/**
 * shallow equal - object and its properties - compare
 * @param objA
 * @param objB
 * @returns
 */
// https://github.com/reduxjs/react-redux/blob/master/src/utils/shallowEqual.js
export function isShallowEqual(objA, objB) {
    if (is(objA, objB))
        return true;
    if (typeof objA !== 'object' || objA === null || typeof objB !== 'object' || objB === null) {
        return false;
    }
    const keysA = Object.keys(objA);
    const keysB = Object.keys(objB);
    if (keysA.length !== keysB.length)
        return false;
    for (let i = 0; i < keysA.length; i++) {
        if (!Object.prototype.hasOwnProperty.call(objB, keysA[i]) ||
            !is(objA[keysA[i]], objB[keysA[i]])) {
            return false;
        }
    }
    return true;
}
export function isEqualForKeys(objA, objB, keys) {
    for (let i = 0; i < keys.length; i++) {
        if (!is(objA[keys[i]], objB[keys[i]])) {
            return false;
        }
    }
    return true;
}
/**
 * deep equal compare
 * @param value1
 * @param value2
 * @returns
 */
export function isDeepEqual(value1, value2) {
    return isEqual(value1, value2);
}
/**
 * use memo instance if value is consider to be not changed during render by shallow equal comparsion
 * @param value
 * @returns
 */
export function useShallowEqualMemo(value) {
    const ref = useRef();
    const isEqual = isShallowEqual(value, ref.current);
    useEffect(() => {
        if (!isEqual) {
            ref.current = value;
        }
    });
    return isEqual ? ref.current : value;
}
/**
 * use memo instance if value is consider to be not changed during render by deep equal comparsion
 * @param value
 * @returns
 */
export function useDeepEqualMemo(value) {
    const ref = useRef();
    const isEqual = isDeepEqual(value, ref.current);
    useEffect(() => {
        if (!isEqual) {
            ref.current = value;
        }
    });
    return isEqual ? ref.current : value;
}
/**
 * component rerender trigger
 */
export function useRerender() {
    const [, rerender] = useReducer((state) => state + 1, 0);
    return rerender;
}
// delay the update by 34 ms(2 frames at 60fps) to render new state
// this gives time for loading cached data and skip
// the component rerender if the state doesn't change
const UPDATE_WAIT = 34;
/**
 * debounced state update/rerender
 * @returns [state, offscreenStateRef, updateState, cancel]
 */
export function useHookStateRef(initialState) {
    const rerender = useRerender();
    const onscreenStateRef = useRef(initialState);
    const offscreenStateRef = useRef(initialState);
    const update = useCallback(debounce(() => {
        if (isDeepEqual(onscreenStateRef.current, offscreenStateRef.current)) {
            return;
        }
        onscreenStateRef.current = offscreenStateRef.current;
        rerender();
    }, UPDATE_WAIT, {
        maxWait: UPDATE_WAIT,
        trailing: true,
    }), []);
    return [
        onscreenStateRef.current,
        offscreenStateRef,
        useCallback((fragment) => {
            offscreenStateRef.current = Object.assign({}, onscreenStateRef.current, offscreenStateRef.current, fragment);
            update();
        }, [update]),
        useCallback(() => {
            update.cancel();
        }, [update]),
    ];
}
