import { CacheMode } from '../types';
import { KeyPrefixHelper } from '../caches/utils';
import { createAbortError, createPromise, } from './utils';
/**
 * SWR request, handles request with cache
 */
export class SWRFetcherRequest {
    constructor(requestControl, cacheKey, options, request, logger) {
        this.requestControl = requestControl;
        this.cacheKey = cacheKey;
        this.request = request;
        this.logger = logger;
        this.isRequestSent = false;
        this.isAborted = false;
        this.cacheControl = new CacheControl(options);
    }
    run() {
        const responseControls = createPromise();
        this.responseResolve = responseControls.resolve;
        this.cacheControl.get(this.cacheKey).then((data) => {
            var _a, _b;
            if (this.isAborted) {
                // might not necessary, just to ensure promise always resolved
                responseControls.resolve(createAbortError());
                return;
            }
            if (data !== undefined) {
                (_a = this.logger) === null || _a === void 0 ? void 0 : _a.call(this, 'valid cache found');
            }
            const response = {
                data,
            };
            // check if cache is fresh and send request if it's not
            if (this.cacheControl.isFresh(this.cacheKey)) {
                // resolve with response no matter it has cache loaded or not,
                // since there is no follow-up steps
                responseControls.resolve(response);
                return;
            }
            (_b = this.logger) === null || _b === void 0 ? void 0 : _b.call(this, 'start fetch from remote');
            let nextResolve;
            if (data !== undefined) {
                const nextControls = createPromise();
                response.next = nextControls.promise;
                // resolve with cache
                responseControls.resolve(response);
                nextResolve = nextControls.resolve;
                // to let abort work on next promise
                this.responseResolve = nextResolve;
            }
            else {
                // no cache, reuse outter response
                nextResolve = responseControls.resolve;
            }
            this.requestControl
                .getResponse(this.cacheKey, this.request)
                .then((data) => {
                var _a;
                this.isRequestSent = false;
                this.cacheControl.set(this.cacheKey, data);
                if (this.isAborted) {
                    // might not necessary, just to ensure promise always resolved
                    nextResolve(createAbortError());
                    return;
                }
                (_a = this.logger) === null || _a === void 0 ? void 0 : _a.call(this, 'fetch from remote success');
                nextResolve({ data });
            })
                .catch((error) => {
                var _a;
                this.isRequestSent = false;
                (_a = this.logger) === null || _a === void 0 ? void 0 : _a.call(this, 'fetch from remote failed');
                nextResolve({ error });
            });
            this.isRequestSent = true;
        });
        return responseControls.promise;
    }
    abort() {
        var _a, _b;
        if (this.isAborted) {
            return;
        }
        this.isAborted = true;
        (_a = this.logger) === null || _a === void 0 ? void 0 : _a.call(this, 'abort called');
        (_b = this.responseResolve) === null || _b === void 0 ? void 0 : _b.call(this, createAbortError());
        if (this.isRequestSent) {
            this.requestControl.release(this.cacheKey);
        }
    }
}
/**
 * cache control helper, handles SWR cache options
 */
class CacheControl {
    constructor(options) {
        this.options = options;
        this.timestampByKey = new Map();
        this.prefixHelper = new KeyPrefixHelper(options.cacheKeyPrefix);
    }
    isFresh(key) {
        switch (this.options.cacheMode) {
            case CacheMode.OnlyIfCached:
                // treat cache as fresh for OnlyIfCached - no matter there is a cache or not
                // so it won't fetch from remote
                return true;
            case CacheMode.NoStore:
            case CacheMode.NoCache:
            case CacheMode.ForceLoad:
                // treat cache as not fresh, so it will always fetch from remote
                return false;
            default:
                break;
        }
        const timestamp = this.timestampByKey.get(key);
        if (!timestamp) {
            return false;
        }
        // cache is always fresh for force cache if there is any cache
        if (this.options.cacheMode === CacheMode.ForceCache) {
            return true;
        }
        return Date.now() - timestamp < this.options.cacheMinFresh * 1000;
    }
    get(key) {
        // force do not use cache
        switch (this.options.cacheMode) {
            case CacheMode.NoStore:
            case CacheMode.NoCache:
                return Promise.resolve(undefined);
            default:
                break;
        }
        return this.options.cache
            .get(this.prefixHelper.appendPrefix(key))
            .then((cachedData) => {
            if (cachedData === undefined) {
                return undefined;
            }
            const { data, timestamp } = cachedData;
            if (!this.timestampByKey.has(key)) {
                this.timestampByKey.set(key, timestamp);
            }
            // force use cache
            switch (this.options.cacheMode) {
                case CacheMode.ForceCache:
                case CacheMode.OnlyIfCached:
                    return data;
                default:
                    break;
            }
            // check cache expires or not
            if (Date.now() - timestamp > this.options.cacheMaxAge * 1000) {
                return undefined;
            }
            return data;
        });
    }
    set(key, value) {
        if (this.options.cacheMode === CacheMode.NoStore) {
            // skip cache for NoStore mode
            return Promise.resolve();
        }
        const timestamp = Date.now();
        this.timestampByKey.set(key, timestamp);
        return this.options.cache.set(this.prefixHelper.appendPrefix(key), {
            data: value,
            timestamp,
        });
    }
}
