import { createPromise } from './requests/utils';
import { CacheMode } from './types';
/**
 * helper for handle each fetcher.fetch() response
 * @param requestReturn the fetcher.fetch() request return
 * @param handler handle that deal with each response
 * @returns abort function that can abort this request
 */
export function forEachResponse(requestReturn, handler) {
    const forEachHandler = (response) => {
        var _a;
        handler(response);
        (_a = response.next) === null || _a === void 0 ? void 0 : _a.then(forEachHandler);
    };
    requestReturn.response.then(forEachHandler);
    return requestReturn.abort;
}
/**
 * helper for handle each fetcher.fetch() response
 * and returns the latest data it receives after all responses settled
 * NOTE:
 *  - it only reject with error if there is no any data available
 *  - this method SHOULD NOT be used together with polling requests - which never ends
 * @param requestReturn the fetcher.fetch() request return
 */
export function getFinalResponse(requestReturn) {
    const { promise, resolve, reject } = createPromise();
    let latestData;
    forEachResponse(requestReturn, ({ data, error, next }) => {
        if (data !== undefined) {
            latestData = data;
        }
        if (!next) {
            if (latestData !== undefined || !error) {
                resolve(latestData);
            }
            else {
                reject(error);
            }
        }
    });
    return promise;
}
/**
 * helper for handle each fetcher.fetch() response
 * and returns the initial data it receives no matter it comes from valid cache or network
 * NOTE:
 *  - it only reject with error if there is no any data available
 *  - this method SHOULD NOT be used together with polling requests - which never ends
 * @param requestReturn the fetcher.fetch() request return
 */
export function getInitialResponse(requestReturn) {
    const { promise, resolve, reject } = createPromise();
    let resolved = false;
    forEachResponse(requestReturn, ({ data, error, next }) => {
        if (data !== undefined && !resolved) {
            resolved = true;
            resolve(data);
            return;
        }
        if (!next && !resolved) {
            if (error) {
                reject(error);
            }
            else {
                resolve(undefined);
            }
        }
    });
    return promise;
}
/**
 * fallback to a pure fetch function, which uses no cache store and return plain promise wrapped response
 * @param fetcher
 */
export function fallbackToPureFetch(fetcher) {
    return (request) => fetcher
        .fetch(request, { cacheMode: CacheMode.NoStore })
        .response.then(({ data, error }) => {
        if (error) {
            throw error;
        }
        return data;
    });
}
/**
 * clear cache
 * @param cache cache to be cleared
 * @param maxAge only keys older than max age will be cleared if maxAge set
 */
export function clearCache(cache, maxAge) {
    return cache.getKeys().then((keys) => Promise.all(keys.map((key) => {
        if (maxAge > 0) {
            return cache.get(key).then((data) => {
                // cache is valid
                if ((data === null || data === void 0 ? void 0 : data.timestamp) + maxAge * 1000 > Date.now()) {
                    return;
                }
                return cache.remove(key);
            });
        }
        return cache.remove(key);
    })));
}
